@@ -55,7 +55,7 @@
 #define A_8R  0x101  /* 8-bit relative address */
 #define A_8X2 0x102  /* Two 8-bit numbers */
 
-/* Argument formating */
+/* Argument formatting */
 #define X_8   "%02xh"
 #define X_8R  "%04xh"
 #define X_16  "%02x%02xh"
@@ -2156,9 +2156,10 @@ int disassemble(int pc)
 	printf (code->name, (pc + 1 + (signed char) mem_read(pc)) & 0xFFFF);
 	break;
     }
+
     putchar ('\n');
 
-    pc += arglen(code->args);
-    return pc;  /* return the location of the next instruction */
+    /* return the location of the next instruction */
+    return pc + arglen(code->args);
 }
 #endif
