@@ -116,24 +116,24 @@ static void m6845_crtc(int value)
       value &= 0x1F;
       m6845.cer = value > m6845.ras ? m6845.ras : value;
       break;
-    case 0x0C: /* Start Address LSB */
+    case 0x0C: /* Start Address MSB */
       m6845.sta = ((value & 0x3F) << 8) | (m6845.sta & 0x00FF);
       return;
-    case 0x0D: /* Start Address MSB */
+    case 0x0D: /* Start Address LSB */
       m6845.sta = ((value & 0xFF) << 0) | (m6845.sta & 0xFF00);
       switch (trs_clones.model) {
         case CT80:
-          mem_video_start(m6845.sta);
+          ct80_video_addr(m6845.sta);
           return;
         case GENIE3S:
           mem_video_page(1024 - m6845.sta);
           return;
       }
       return;
-    case 0x0E: /* Cursor LSB */
+    case 0x0E: /* Cursor MSB */
       m6845.pos = ((value & 0x3F) << 8) | (m6845.pos & 0x00FF);
       return;
-    case 0x0F: /* Cursor MSB */
+    case 0x0F: /* Cursor LSB */
       m6845.pos = ((value & 0xFF) << 0) | (m6845.pos & 0xFF00);
       if (m6845.pos != m6845.old) {
         if (m6845.vis)
@@ -148,6 +148,22 @@ static void m6845_crtc(int value)
   m6845_cursor(m6845.pos - m6845.sta, m6845.csr, m6845.cer, m6845.vis);
 }
 
+static int m6845_read(void)
+{
+  switch (m6845_reg) {
+    case 0x0A: /* Cursor Start */
+      return m6845.csr;
+    case 0x0B: /* Cursor End */
+      return m6845.cer;
+    case 0x0E: /* Cursor MSB */
+      return (m6845.pos >> 8) & 0xFF;
+    case 0x0F: /* Cursor LSB */
+      return (m6845.pos >> 0) & 0xFF;
+    default:
+      return 0xFF;
+  }
+}
+
 static int rtc_read(int port)
 {
   const time_t time_secs = time(NULL) + trs_timeoffset;
@@ -229,7 +245,7 @@ void z80_out(int port, int value)
       case EG3200:
         switch (port) {
           case 0x28: /* Genieplus Memory Card */
-            mem_card_out(GENIEPLUS, value);
+            eg3200_genieplus_out(value);
             break;
           case 0x48: /* TRS_HARD_DATA */
           case 0x49: /* TRS_HARD_PRECOMP */
@@ -266,7 +282,7 @@ void z80_out(int port, int value)
             m6845_crtc(value);
             break;
           case 0xFA:
-            eg3200_bank_out(value);
+            sys_byte_write(value & 0x0F);
             break;
           case 0xFD:
             trs_printer_write(value);
@@ -316,6 +332,7 @@ void z80_out(int port, int value)
             trs_disk_data_write(value);
             break;
           case 0xF1:
+            /* Save SIO for GENIE-DOS 2.4 */
             modeimage = value;
             break;
           case 0xF6:
@@ -331,7 +348,7 @@ void z80_out(int port, int value)
             genie3s_sys_out(value);
             break;
           case 0xFB:
-            mem_card_out(TCS_GENIE3S, value);
+            genie3s_mem_out(value);
             break;
           case 0xFD:
             trs_printer_write(value);
@@ -350,7 +367,6 @@ void z80_out(int port, int value)
       case LNW80:
         switch (port) {
           case 0x1F:
-          case 0x3F:
             lnw80_bank_out(value);
             return;
           case 0xFE:
@@ -373,7 +389,7 @@ void z80_out(int port, int value)
             rtc_reg = value >> 4;
             break;
           case 0x7E:
-            mem_card_out(RAM192B, value);
+            tcs_ram192b_out(value);
             break;
           case 0xFE:
             sys_byte_out(value);
@@ -408,8 +424,17 @@ void z80_out(int port, int value)
           case 0x1F:
           case 0x3F:
             if (selector)
-              mem_card_out(SELECTOR, value);
+              selector_out(value);
             return;
+          /* X-MEM/80 */
+          case 0x41: /* lower 16K 8000-BFFF */
+          case 0x42: /* upper 16K C000-FFFF */
+          case 0x43: /* both  16K same page */
+            if (xmem80) {
+              xmem80_out(port, value & 0x1F);
+              return;
+            }
+            break;
           case 0x85: /* Real Time Clock/Calendar Card */
             rtc_reg = value;
             return;
@@ -421,7 +446,7 @@ void z80_out(int port, int value)
             return;
           case 0xC0: /* Lubomir Soft Banker or EG 64.3 */
             if (lubomir)
-              mem_card_out(LUBOMIR, value);
+              lubomir_out(value);
             return;
           case 0xD0: /* Schmidtke 80-Z Card: 6845 CRTC */
           case 0x10: /* Homebrew 80*22 SYS80.SYS */
@@ -501,6 +526,18 @@ void z80_out(int port, int value)
   } else {
     /* Next, Models III/4/4P only */
     switch (port) {
+      case 0x3C:
+        if (trs_model == 3)
+          m6845_reg = value;
+        return;
+      case 0x3D:
+        if (trs_model == 3)
+          m6845_crtc(value);
+        return;
+      case 0x3F:
+        if (trs_model == 3)
+          vid80_vx3_out(value);
+        return;
       case 0x50: /* MegaMem memory slot */
       case 0x51:
       case 0x52:
@@ -574,7 +611,7 @@ void z80_out(int port, int value)
         /* HyperMem uses bits 4-1 of this port, 0 is the existing
            sound */
         if (trs_model >= 4 && hypermem)
-          mem_card_out(HYPERMEM, value);
+          hypermem_out(value);
         /* Fall through - we affect the sound as well */
       case 0x91:
       case 0x92:
@@ -583,14 +620,14 @@ void z80_out(int port, int value)
         return;
       case 0x94: /* Huffman memory expansion */
         if (trs_model >= 4 && huffman)
-          mem_card_out(HUFFMAN, value);
+          huffman_out(value);
         return;
       case 0x9C:
       case 0x9D:
       case 0x9E:
       case 0x9F:
-        if (trs_model == 5 /*4P*/)
-          mem_romin(value & 1);
+        if (trs_model == 5) /* Switch 4P boot ROM in/out */
+          sys_byte_write(value & 1);
         return;
       case 0xE0:
       case 0xE1:
@@ -675,8 +712,8 @@ void z80_out(int port, int value)
   /* Ports common to all TRS-80 models */
   switch (port) {
     case 0x43: /* Alpha Technology SuperMem */
-      if (supermem)
-        mem_card_out(SUPERMEM, value);
+      if ((value & 0x1F) < supermem)
+        supermem_out(value);
       break;
     case TRS_HARD_WP:       /* 0xC0 */
     case TRS_HARD_CONTROL:  /* 0xC1 */
@@ -723,9 +760,6 @@ int z80_in(int port)
           case 0xD7:
             value = ct80_ramdisk_in(port);
             goto done;
-          case 0xFE:
-            value = sys_byte_in();
-            goto done;
           case 0xFF:
             value = (modesel ? 0x3F : 0x7F) | trs_cassette_in();
             goto done;
@@ -756,17 +790,7 @@ int z80_in(int port)
             value = rtc_read(rtc_reg);
             break;
           case 0xF7:
-            switch (m6845_reg) {
-              case 0x0E: /* Cursor LSB */
-                value = (m6845.pos >> 8) & 0xFF;
-                break;
-              case 0x0F: /* Cursor MSB */
-                value = (m6845.pos >> 0) & 0xFF;
-                break;
-            }
-            break;
-          case 0xFA:
-            value = sys_byte_in();
+            value = m6845_read();
             break;
           case 0xFD:
             value = trs_printer_read();
@@ -816,17 +840,11 @@ int z80_in(int port)
             value = trs_disk_data_read();
             break;
           case 0xF1:
+            /* Saved SIO baud rate */
             value = modeimage;
             break;
           case 0xF7:
-            switch (m6845_reg) {
-              case 0x0E: /* Cursor LSB */
-                value = (m6845.pos >> 8) & 0xFF;
-                break;
-              case 0x0F: /* Cursor MSB */
-                value = (m6845.pos >> 0) & 0xFF;
-                break;
-            }
+            value = m6845_read();
             break;
           case 0xF9:
             value = sys_byte_2_in();
@@ -835,7 +853,7 @@ int z80_in(int port)
             value = sys_byte_in();
             break;
           case 0xFB:
-            value = mem_card_in(TCS_GENIE3S);
+            value = genie3s_mem_in();
             break;
           case 0xFD:
             value = trs_printer_read();
@@ -868,7 +886,7 @@ int z80_in(int port)
             value = rtc_read(rtc_reg);
             break;
           case 0x7E:
-            value = mem_card_in(RAM192B);
+            value = sys_byte_2_in();
             break;
           case 0xFE:
             value = sys_byte_in();
@@ -893,6 +911,15 @@ int z80_in(int port)
           case 0x04: /* HRG read data byte */
             value = hrg_read_data();
             goto done;
+          /* X-MEM/80 */
+          case 0x41: /* lower 16K 8000-BFFF */
+          case 0x42: /* upper 16K C000-FFFF */
+          case 0x43: /* both  16K same page */
+            if (xmem80) {
+              value = xmem80_in(port);
+              goto done;
+            }
+            break;
           case 0x84: /* RTCC */
             value = rtc_read(rtc_reg);
             goto done;
@@ -929,12 +956,20 @@ int z80_in(int port)
   } else {
     /* Models III/4/4P only */
     switch (port) {
+      case 0x3D:
+        if (trs_model == 3)
+          value = m6845_read();
+        goto done;
+      case 0x3F:
+        if (trs_model == 3)
+          value = sys_byte_in();
+        goto done;
       case 0x82:
         value = grafyx_read_data();
         goto done;
       case 0x94: /* Huffman memory expansion */
         if (trs_model >= 4 && huffman)
-          value = mem_card_in(HUFFMAN);
+          value = huffman_in();
         goto done;
       case 0xE0:
       case 0xE1:
@@ -970,7 +1005,8 @@ int z80_in(int port)
         value = trs_disk_data_read();
         goto done;
       case 0xF4:
-        value = cp500_switch_mode(Z80_A);
+        if (trs_model == 3)
+          value = cp500_switch_mode(Z80_A);
         goto done;
       case 0xF8:
       case 0xF9:
@@ -993,7 +1029,7 @@ int z80_in(int port)
       goto done;
     case 0x43: /* Supermem memory expansion */
       if (supermem)
-        value = mem_card_in(SUPERMEM);
+        value = supermem_in();
       goto done;
     case TRS_HARD_WP:       /* 0xC0 */
     case TRS_HARD_CONTROL:  /* 0xC1 */
@@ -1076,7 +1112,7 @@ int z80_in(int port)
 done:
 #ifdef ZBX
   if (trs_io_debug_flags & IODEBUG_IN)
-    debug("[PC=%04X] in (0x%02X) => 0x%02X\n", Z80_PC, port, value);
+    debug("[PC=%04X] in (0x%02X) = 0x%02X\n", Z80_PC, port, value);
 #endif
 
   return value;
