@@ -228,10 +228,10 @@ static const struct {
   { "hd",              opt_dirname,       1, 0, trs_hard_dir          },
   { "hdboot",          opt_value,         0, 1, &trs_hd_boot          },
   { "hideled",         opt_value,         0, 0, &trs_show_led         },
-  { "huff",            opt_memory,        0, 2, &huffman              },
-  { "huffman",         opt_memory,        0, 2, &huffman              },
-  { "hyper",           opt_memory,        0, 3, &hypermem             },
-  { "hypermem",        opt_memory,        0, 3, &hypermem             },
+  { "huff",            opt_memory,        0, 1, &huffman              },
+  { "huffman",         opt_memory,        0, 1, &huffman              },
+  { "hyper",           opt_memory,        0, 2, &hypermem             },
+  { "hypermem",        opt_memory,        0, 2, &hypermem             },
   { "joyaxismapped",   opt_value,         0, 1, &jaxis_mapped         },
   { "joybuttonmap",    opt_joybuttonmap,  1, 0, NULL                  },
   { "joystick",        opt_joystick,      1, 0, NULL                  },
@@ -246,16 +246,16 @@ static const struct {
   { "led",             opt_value,         0, 1, &trs_show_led         },
   { "lower",           opt_value,         0, 1, &lowercase            },
   { "lowercase",       opt_value,         0, 1, &lowercase            },
-  { "lsb",             opt_memory,        0, 4, &lubomir              },
-  { "lubomir",         opt_memory,        0, 4, &lubomir              },
+  { "lsb",             opt_memory,        0, 3, &lubomir              },
+  { "lubomir",         opt_memory,        0, 3, &lubomir              },
   { "micro",           opt_value,         0, 1, &grafyx_microlabs     },
   { "microlabs",       opt_value,         0, 1, &grafyx_microlabs     },
   { "m1",              opt_value,         0, 1, &trs_model            },
   { "m3",              opt_value,         0, 3, &trs_model            },
   { "m4",              opt_value,         0, 4, &trs_model            },
   { "m4p",             opt_value,         0, 5, &trs_model            },
-  { "mega",            opt_memory,        0, 5, &megamem              },
-  { "megamem",         opt_memory,        0, 5, &megamem              },
+  { "mega",            opt_memory,        0, 4, &megamem              },
+  { "megamem",         opt_memory,        0, 4, &megamem              },
   { "m",               opt_model,         1, 0, NULL                  },
   { "model",           opt_model,         1, 0, NULL                  },
   { "mouse",           opt_value,         0, 1, &mousepointer         },
@@ -307,6 +307,8 @@ static const struct {
   { "not",             opt_value,         0, 0, &turbo_mode           },
   { "notruedam",       opt_value,         0, 0, &trs_disk_truedam     },
   { "noturbo",         opt_value,         0, 0, &turbo_mode           },
+  { "noxmem",          opt_value,         0, 0, &xmem80               },
+  { "noxmem80",        opt_value,         0, 0, &xmem80               },
   { "pause",           opt_value,         0, 1, &trs_paused           },
   { "p",               opt_printer,       1, 0, NULL                  },
   { "pd",              opt_dirname,       1, 0, trs_printer_dir       },
@@ -338,8 +340,8 @@ static const struct {
   { "scanlines",       opt_value,         0, 1, &scanlines            },
   { "scanshade",       opt_intval,        1, 5, NULL                  },
   { "sd",              opt_dirname,       1, 0, trs_state_dir         },
-  { "sel",             opt_memory,        0, 7, &selector             },
-  { "selector",        opt_memory,        0, 7, &selector             },
+  { "sel",             opt_memory,        0, 5, &selector             },
+  { "selector",        opt_memory,        0, 5, &selector             },
   { "serial",          opt_serial,        1, 0, NULL                  },
   { "shade",           opt_intval,        1, 5, NULL                  },
   { "shiftbracket",    opt_shiftbracket,  0, 1, NULL                  },
@@ -351,8 +353,8 @@ static const struct {
   { "sr",              opt_intval,        1, 3, NULL                  },
   { "statedir",        opt_dirname,       1, 0, trs_state_dir         },
   { "stringy",         opt_value,         0, 1, &stringy              },
-  { "super",           opt_memory,        0, 8, &supermem             },
-  { "supermem",        opt_memory,        0, 8, &supermem             },
+  { "super",           opt_intval,        1, 8, NULL                  },
+  { "supermem",        opt_intval,        1, 8, NULL                  },
   { "sw",              opt_switches,      1, 0, &trs_uart_switches    },
   { "switches",        opt_switches,      1, 0, &trs_uart_switches    },
   { "t",               opt_value,         0, 1, &turbo_mode           },
@@ -382,6 +384,8 @@ static const struct {
   { "wafer5",          opt_wafer,         1, 5, NULL                  },
   { "wafer6",          opt_wafer,         1, 6, NULL                  },
   { "wafer7",          opt_wafer,         1, 7, NULL                  },
+  { "xmem",            opt_memory,        0, 7, &xmem80               },
+  { "xmem80",          opt_memory,        0, 7, &xmem80               },
   { "y",               opt_intval,        1, 7, NULL                  },
   { "year",            opt_intval,        1, 7, NULL                  },
 };
@@ -665,6 +669,14 @@ static void opt_intval(const char *arg, int intarg, int *stringarg)
       if (value >= 0)
         trs_year = value;
       break;
+    case 8: /* supermem */
+      if (value >= 256 && value <= 1024) {
+        supermem = (value / 256) << 3;
+         huffman = hypermem = lubomir = selector = xmem80 = 0;
+      } else {
+        supermem = 0;
+      }
+      break;
     default:
       break;
   }
@@ -713,6 +725,7 @@ static void opt_memory(const char *arg, int intarg, int *variable)
     case LUBOMIR:
       selector = 0;
       supermem = 0;
+      xmem80   = 0;
       break;
     case MEGAMEM:
       huffman  = 0;
@@ -720,12 +733,12 @@ static void opt_memory(const char *arg, int intarg, int *variable)
     case SELECTOR:
       lubomir  = 0;
       supermem = 0;
+      xmem80   = 0;
       break;
-    case SUPERMEM:
+    case XMEM80:
       lubomir  = 0;
-      huffman  = 0;
-      hypermem = 0;
       selector = 0;
+      supermem = 0;
       break;
   }
 }
@@ -767,10 +780,14 @@ static void opt_printer(const char *arg, int intarg, int *stringarg)
     case '0':
       trs_printer = 0;
       break;
-    case 't': /* text */
+    case 'f': /* file */
     case '1':
       trs_printer = 1;
       break;
+    case 't': /* text */
+    case '2':
+      trs_printer = 2;
+      break;
     default:
       error("unknown printer type: '%s'", arg);
   }
@@ -970,6 +987,7 @@ int trs_load_config_file(void)
 #if defined(SDL2) || !defined(NOX)
   turbo_paste = 0;
 #endif
+  xmem80 = 0;
   Z80_HALT = 0;
 
   if (trs_config_file[0] == 0) {
@@ -1161,7 +1179,11 @@ int trs_write_config_file(const char *filename)
       trs_model == 5 ? 4 : trs_model, trs_model == 5 ? "P" : "");
 
   fprintf(config_file, "%smousepointer\n", option(mousepointer));
-  fprintf(config_file, "printer=%s\n", trs_printer ? "text" : "none");
+
+  fprintf(config_file, "printer=%s\n",
+      trs_printer == 1 ? "file" :
+      trs_printer == 2 ? "text" : "none");
+
   fprintf(config_file, "printerdir=%s\n", trs_printer_dir);
   fprintf(config_file, "%sresize3\n", option(resize3));
   fprintf(config_file, "%sresize4\n", option(resize4));
@@ -1199,7 +1221,7 @@ int trs_write_config_file(const char *filename)
       trs_disk_getstep(4), trs_disk_getstep(5), trs_disk_getstep(6), trs_disk_getstep(7));
 #endif
   fprintf(config_file, "%sstringy\n", option(stringy));
-  fprintf(config_file, "%ssupermem\n", option(supermem));
+  fprintf(config_file, "supermem=%d\n", supermem << 5);
   fprintf(config_file, "switches=0x%x\n", trs_uart_switches);
   fprintf(config_file, "%struedam\n", option(trs_disk_truedam));
   fprintf(config_file, "%sturbo\n", option(turbo_mode));
@@ -1211,6 +1233,7 @@ int trs_write_config_file(const char *filename)
   for (i = 0; i < 8; i++)
     fprintf(config_file, "wafer%d=%s\n", i, stringy_get_name(i));
 
+  fprintf(config_file, "%sxmem80\n", option(xmem80));
   fprintf(config_file, "year=%d\n", trs_year);
 
   fclose(config_file);
