@@ -45,6 +45,7 @@
 #include "trs_cassette.h"
 #include "trs_disk.h"
 #include "trs_hard.h"
+#include "trs_memory.h"
 #include "trs_mkdisk.h"
 #include "trs_sdl_gui.h"
 #include "trs_sdl_keyboard.h"
@@ -144,7 +145,7 @@ static void gui_delete_filelist(void);
 static void gui_add_to_filelist(char *name);
 static int  gui_filename_cmp(const void *nptr1, const void *nptr2);
 static int  gui_read_dir(const char *path, const char *mask, int dir);
-static int  gui_input(const char *title, const char *input, char *output, int limit, int file);
+static int  gui_input(const char *title, const char *input, char *output, int limit, int dir);
 static int  gui_menu(const char *title, const MENU *entry, int selection);
 static int  gui_popup(const char *title, const char **entry, int num, int selection);
 static int  gui_matrix(const char *title, const char **entry, int rows, int cols, int selection);
@@ -248,12 +249,6 @@ int gui_key(void)
 {
   SDL_Event event;
 
-#ifdef SDL2
-  /* Stop Text input first to prevent double chars */
-  SDL_StopTextInput();
-  SDL_StartTextInput();
-#endif
-
   while (1) {
     SDL_WaitEvent(&event);
     switch (event.type) {
@@ -287,7 +282,6 @@ int gui_key(void)
         break;
 
       case SDL_TEXTINPUT:
-        SDL_StopTextInput();
         return event.text.text[0];
 
       case SDL_WINDOWEVENT:
@@ -302,7 +296,7 @@ int gui_key(void)
 #endif
             case SDLK_q:
             case SDLK_END:
-              trs_exit(1);
+              trs_exit(!(SDL_GetModState() & KMOD_SHIFT));
               break;
             case SDLK_BACKSPACE:
               return SDLK_F9;
@@ -763,9 +757,9 @@ done:
   return selection;
 }
 
-int gui_input(const char *title, const char *input, char *output, int limit, int file)
+int gui_input(const char *title, const char *input, char *output, int limit, int dir)
 {
-  int insert = 1;
+  int ins = 1;
   int pos;
   int len;
   int col;
@@ -779,10 +773,12 @@ redraw:
   col = pos > 60 ? pos - 59 : 0;
 
   gui_rect(1, 6, 62, 3, 1);
-  gui_text(insert ? " INS " : " OVR ", 56, 8, 6, 5);
   gui_write(title, 0, 6, 4);
 
-  if (file)
+  if (ins)
+    gui_text(" INS ", 56, 8, 6, 5);
+
+  if (dir)
     gui_write(" TAB:Select Directory ", 0, 8, 5);
 
   while (1) {
@@ -817,7 +813,8 @@ redraw:
       case SDLK_END:
       case SDLK_PAGEDOWN:
         pos = len;
-        goto redraw;
+        col = pos > 60 ? pos - 59 : 0;
+        break;
       case SDLK_BACKSPACE:
         if (pos > 0) {
           for (i = pos; i < len; i++)
@@ -838,7 +835,7 @@ redraw:
         }
         break;
       case SDLK_INSERT:
-        insert = !insert;
+        ins = !ins;
         goto redraw;
       case SDLK_RETURN:
         return len;
@@ -847,7 +844,7 @@ redraw:
       case SDLK_DOWN:
       case SDLK_TAB:
       case SDLK_UP:
-        if (file) {
+        if (dir) {
           char directory[FILENAME_MAX];
 
           if (gui_file(input, directory, NULL, 1, "") >= 0) {
@@ -865,7 +862,7 @@ redraw:
         break;
       default:
         if (key >= ' ' && key <= 0xFF && pos < limit) {
-          if (insert && len < limit) {
+          if (ins && len < limit) {
             for (i = len; i > pos; i--)
               output[i] = output[i - 1];
             len++;
@@ -1014,8 +1011,10 @@ int gui_menu(const char *title, const MENU *entry, int selection)
           }
           return selection;
         case SDLK_SPACE:
-          if (trs_write_protect(entry[selection].type, selection) < 0)
-            gui_error("Write-Protect");
+          if (entry[selection].type < ENTRY) {
+            if (trs_write_protect(entry[selection].type, selection) < 0)
+              gui_error("Write-Protection");
+          }
           return selection;
         case SDLK_BACKSPACE:
         case SDLK_ESCAPE:
@@ -1357,10 +1356,10 @@ void gui_disk_options(void)
     gui_clear();
 
     selection = gui_menu(" Floppy Disk Options ", menu, selection);
-    if (selection >= 0 && selection < 8) {
-      trs_disk_setsize(selection, (trs_disk_getsize(selection) == 8) ? 5 : 8);
-    }
-    else switch (selection) {
+    switch (selection) {
+      case 0: case 1: case 2: case 3: case 4: case 5: case 6: case 7:
+        trs_disk_setsize(selection, (trs_disk_getsize(selection) == 8) ? 5 : 8);
+        break;
       case 9:
         trs_disk_controller = !trs_disk_controller;
         break;
@@ -1478,7 +1477,7 @@ void gui_hard_menu(void)
       menu[i].text[0] = trs_hard_getwriteprotect(i) ? '*' : ' ';
     }
 
-    if (selection >= 0 && selection < 4)
+    if (selection < 4)
       trs_hard_getgeometry(selection, &cylinders, &heads, &sectors);
 
     snprintf(&menu[8].text[55], 6, "%5d", cylinders);
@@ -1708,6 +1707,7 @@ void gui_emulator_menu(void)
    {"Lowercase Modification for Model I                ", ENTRY},
    {"Lubomir Soft Banker Emulation for Model I         ", ENTRY},
    {"Selector (Dutch TRS-80 Users Society) Model I     ", ENTRY},
+   {"X-MEM/80 16K page (Michael Wessel) for Model I    ", ENTRY},
    {"SuperMem (Alpha Technology) I/III/4/4P Memory     ", ENTRY},
    {"Grafyx Solution (Micro-Labs) III/4/4P Graphics    ", ENTRY},
    {"MegaMem (Anitek Software) III/4/4P Memory Board   ", ENTRY},
@@ -1726,6 +1726,11 @@ void gui_emulator_menu(void)
                          "    CT-80 Aster",
                          "       LNW80/II",
                          "TCS SpeedMaster"};
+  const char *super[] = {"        No",
+                         "    256 KB",
+                         "    512 KB",
+                         "    768 KB",
+                         "   1024 KB"};
   int selection = 0;
   int model_selection = trs_model == 1 ? 0 : trs_model - 2;
   float clock_mhz[4];
@@ -1746,11 +1751,12 @@ void gui_emulator_menu(void)
     snprintf(&menu[5].text[50], 11, "%s", yes_no[lowercase]);
     snprintf(&menu[6].text[50], 11, "%s", yes_no[lubomir]);
     snprintf(&menu[7].text[50], 11, "%s", yes_no[selector]);
-    snprintf(&menu[8].text[50], 11, "%s", yes_no[supermem]);
-    snprintf(&menu[9].text[50], 11, "%s", yes_no[grafyx_microlabs]);
-    snprintf(&menu[10].text[50], 11, "%s", yes_no[megamem]);
-    snprintf(&menu[11].text[50], 11, "%s", yes_no[huffman]);
-    snprintf(&menu[12].text[50], 11, "%s", yes_no[hypermem]);
+    snprintf(&menu[8].text[50], 11, "%s", yes_no[xmem80]);
+    snprintf(&menu[9].text[50], 11, "%s", super[supermem >> 3]);
+    snprintf(&menu[10].text[50], 11, "%s", yes_no[grafyx_microlabs]);
+    snprintf(&menu[11].text[50], 11, "%s", yes_no[megamem]);
+    snprintf(&menu[12].text[50], 11, "%s", yes_no[huffman]);
+    snprintf(&menu[13].text[50], 11, "%s", yes_no[hypermem]);
     gui_clear();
 
     selection = gui_menu(" Emulator Settings ", menu, selection);
@@ -1798,32 +1804,37 @@ void gui_emulator_menu(void)
       case 6:
         lubomir = !lubomir;
         if (lubomir)
-          selector = supermem = 0;
+          selector = supermem = xmem80 = 0;
         break;
       case 7:
         selector = !selector;
         if (selector)
-          lubomir = supermem = 0;
+          lubomir = supermem = xmem80 = 0;
         break;
       case 8:
-        supermem = !supermem;
-        if (supermem)
-          huffman = hypermem = lubomir = selector = 0;
+        xmem80 = !xmem80;
+        if (xmem80)
+          lubomir = selector = supermem = 0;
         break;
       case 9:
-        grafyx_microlabs = !grafyx_microlabs;
+        supermem = gui_popup("SuperMem", super, 5, supermem >> 3) << 3;
+        if (supermem)
+          huffman = hypermem = lubomir = selector = xmem80 = 0;
         break;
       case 10:
+        grafyx_microlabs = !grafyx_microlabs;
+        break;
+      case 11:
         megamem = !megamem;
         if (megamem)
           huffman = 0;
         break;
-      case 11:
+      case 12:
         huffman = !huffman;
         if (huffman)
           megamem = hypermem = supermem = 0;
         break;
-      case 12:
+      case 13:
         hypermem = !hypermem;
         if (hypermem)
           huffman = supermem = 0;
@@ -1973,9 +1984,7 @@ void gui_display_menu(void)
         if (gui_input(" Enter Window border width in pixels (0 to 50) ",
             input, input, 2, 0) > 0) {
           value = atoi(input);
-          if (value != border_width) {
-            if (value < 0 || value > 50)
-              value = 2;
+          if (value != border_width && value >= 0 && value <= 50) {
             border_width = value;
             resize = 1;
           }
@@ -2048,7 +2057,7 @@ void gui_misc_menu(void)
    {"Turbo Paste                                             ", ENTRY},
 #endif
    {"", 0}};
-  const char *printer[] = {"     None", "     Text"};
+  const char *printer[] = {"     None", "     File", "     Text"};
   int selection = 0;
 
   while (1) {
@@ -2099,7 +2108,7 @@ void gui_misc_menu(void)
         }
         break;
       case 4:
-        trs_printer = !trs_printer;
+        trs_printer = gui_popup("Printer", printer, 3, trs_printer);
         break;
       case 6:
         filename[0] = 0;
@@ -2603,7 +2612,7 @@ void gui_about(void)
   gui_clear();
   gui_write(" About ", 0, 0, 4);
   gui_write("SDLTRS", 0, 3, 4);
-  gui_write("Version 1.2.33", 0, 4, 4);
+  gui_write("Version 1.2.34", 0, 4, 4);
   gui_write("BSD 2-Clause License", 0, 5, 4);
   gui_write("Copyright (C) 2006-2011 Mark Grebe, 2018-2025", 0, 6, 4);
   gui_write("Alan Cox, Jens Guenther, Leonardo Brondani Schenkel", 0, 7, 4);
@@ -2757,6 +2766,11 @@ void gui_main_menu(void)
 
 void gui_function(int function)
 {
+#ifdef SDL2
+  /* Prevent double chars */
+  SDL_FlushEvent(SDL_TEXTINPUT);
+#endif
+
   SDL_PauseAudio(1);
 
   switch (function) {
