@@ -60,6 +60,8 @@
 #define TRS_CHAR_HEIGHT  12
 #define TRS_CHAR_HEIGHT4 10
 
+int text80x24;
+
 /* Private data */
 #include "trs_chars.c"
 
@@ -74,7 +76,7 @@ static int m6845_raster = 12;
 static int window_scale;
 static int y_scale = 2;
 static int trs_charset;
-static int text80x24, screen640x240;
+static int screen640x240;
 static int drawnRectCount;
 static int top_margin;
 static int left_margin;
@@ -502,13 +504,9 @@ static void trs_screen_print(void)
   int pos;
 
   for (pos = 0; pos < screen_chars ;) {
-    int byte = trs_screen[pos] & 0x7F;
-
-    if (byte < ' ') byte = '.';
-    trs_printer_write(byte);
-
+    trs_printer_write(trs_screen[pos]);
     if ((++pos) % cols == 0)
-      trs_printer_write('\n');
+      trs_printer_write(0x0D);
   }
 }
 
@@ -575,8 +573,8 @@ void trs_screen_init(int resize)
 
   y_scale = scale * scale_factor;
 
-  char_height = TRS_CHAR_HEIGHT * y_scale;
-  char_width  = TRS_CHAR_WIDTH  * scale;
+  char_height = m6845_raster * y_scale;
+  char_width  = TRS_CHAR_WIDTH * scale;
 
   if (trs_model == 1) {
     switch (trs_clones.model) {
@@ -587,8 +585,9 @@ void trs_screen_init(int resize)
         /* Use alternate font for Holte-ROM */
         trs_charset = trs_rom_size > 2048 ? 16 : 15;
         break;
-      case GENIE3S:
-        char_height = m6845_raster * y_scale;
+      case LNW80:
+        if (cols == 80) char_height = 10 * y_scale;
+        trs_charset = trs_charset1;
         break;
       case SPEEDMASTER:
         trs_charset = 17;
@@ -655,6 +654,9 @@ void trs_screen_init(int resize)
     if (screen == NULL)
       fatal("SDL_SetVideoMode failed: %s", SDL_GetError());
 #endif
+    screen_chars = cols * rows;
+    if (screen_chars > SCREEN_SIZE)
+      screen_chars = SCREEN_SIZE;
   }
 
 #if defined(big_endian) && !defined(__linux__)
@@ -1036,7 +1038,7 @@ void trs_exit(int confirm)
   exit(EXIT_SUCCESS);
 }
 
-static void trs_set_fullscreen(int on_off)
+static void trs_fullscreen(int on_off)
 {
   if (fullscreen == on_off) return;
   fullscreen = on_off;
@@ -1061,8 +1063,6 @@ void trs_get_event(int wait)
   SDL_Event event;
 #ifdef SDL2
   SDL_Keysym keysym;
-
-  SDL_StartTextInput();
 #else
   SDL_keysym keysym;
 #endif
@@ -1181,10 +1181,10 @@ void trs_get_event(int wait)
               trs_screen_caption();
             } else {
 #ifdef ZBX
-              trs_set_fullscreen(0);
+              trs_fullscreen(0);
               trs_debug();
 #else
-              trs_set_fullscreen(!fullscreen);
+              trs_fullscreen(!fullscreen);
 #endif
             }
             continue;
@@ -1257,6 +1257,9 @@ void trs_get_event(int wait)
               SDL_ShowCursor(SDL_ENABLE);
               break;
 #endif
+            case SDLK_BACKSPACE:
+              trs_timer_init();
+              break;
             case SDLK_DELETE:
               trs_reset(SDL_GetModState() & KMOD_SHIFT);
               break;
@@ -1264,24 +1267,24 @@ void trs_get_event(int wait)
               gui_function(KEYBRD);
               break;
             case SDLK_RETURN:
-              trs_set_fullscreen(!fullscreen);
+              trs_fullscreen(!fullscreen);
               break;
             case SDLK_HOME:
-              trs_set_fullscreen(0);
+              trs_fullscreen(0);
               if (scale != 1) {
                 scale = 1;
                 trs_screen_init(1);
               }
               break;
             case SDLK_PAGEDOWN:
-              trs_set_fullscreen(0);
+              trs_fullscreen(0);
               if (scale < MAX_SCALE) {
                 scale++;
                 trs_screen_init(1);
               }
               break;
             case SDLK_PAGEUP:
-              trs_set_fullscreen(0);
+              trs_fullscreen(0);
               if (scale > 1) {
                 scale--;
                 trs_screen_init(1);
@@ -1358,7 +1361,7 @@ void trs_get_event(int wait)
 #endif
             case SDLK_q:
             case SDLK_END:
-              trs_exit(1);
+              trs_exit(!(SDL_GetModState() & KMOD_SHIFT));
               break;
             case SDLK_r:
               gui_function(READ);
@@ -1385,10 +1388,10 @@ void trs_get_event(int wait)
               break;
             case SDLK_z:
 #ifdef ZBX
-              trs_set_fullscreen(0);
+              trs_fullscreen(0);
               trs_debug();
 #else
-              trs_set_fullscreen(!fullscreen);
+              trs_fullscreen(!fullscreen);
 #endif
               break;
             default:
@@ -1510,7 +1513,10 @@ void trs_get_event(int wait)
           }
         } else {
           if (keysym.sym == SDLK_LCTRL && trs_model == 3) {
-            keysym.sym = 0x117; /* Shifted Down Arrow */
+            if (trs_clones.model == CP500_M80)
+              keysym.sym = 0x117; /* Shifted Down Arrow */
+            else
+              keysym.sym = 0x7F;  /* Clear */
             goto done;
           }
         }
@@ -1747,9 +1753,6 @@ done:
       gui_pause();
 
   } while (!wait);
-#ifdef SDL2
-  SDL_StopTextInput();
-#endif
 }
 
 void trs_screen_mode(int mode, int flag)
@@ -1769,15 +1772,11 @@ static void trs_screen_640x240(int flag)
   if (flag) {
     cols = 80;
     rows = 24;
-    char_height = TRS_CHAR_HEIGHT4 * y_scale;
   } else {
     cols = 64;
     rows = 16;
-    char_height = TRS_CHAR_HEIGHT * y_scale;
   }
 
-  screen_chars = cols * rows;
-
   trs_screen_init(1);
 }
 
@@ -1787,10 +1786,10 @@ void trs_screen_80x24(int flag)
 
   text80x24 = flag;
 
-  if (!grafyx_enable || grafyx_overlay)
+  if (!grafyx_enable || grafyx_overlay) {
+    memset(trs_screen, ' ', SCREEN_SIZE);
     trs_screen_640x240(flag);
-
-  memset(trs_screen, ' ', SCREEN_SIZE);
+  }
 }
 
 static void
@@ -2013,8 +2012,8 @@ void trs_screen_refresh(void)
   } else {
     int i;
 
-    /* Redraw HRG screen */
-    if (hrg_enable) {
+    /* Redraw 480*192 HRG */
+    if (hrg_enable == 2) {
       memset(grafyx_unscaled, 0, G_YSIZE * G_XSIZE);
 
       grafyx_overlay = 0;
@@ -2048,6 +2047,12 @@ void trs_screen_refresh(void)
         z80_out(0xFD, 0xFF);
       }
       break;
+    case 0x34: /* Holmes VID-80 */
+      if (cols == 80) {
+        z80_out(0x3C, 0x09);
+        z80_out(0x3D, 0xFF);
+      }
+      break;
     default:
       break;
   }
@@ -2198,40 +2203,30 @@ void trs_screen_write_char(unsigned int position, Uint8 character)
   if (expanded)
     srcRect.w *= 2;
 
-  switch (trs_clones.model) {
-    case CT80:
-    case GENIE3S:
-      if (grafyx_overlay == 0) goto overlay;
-      break;
-    case LNW80:
-      if (rows == 16) {
-        if (currentmode == INVERSE) expanded += 2;
-        if (character >= 0x80 && character <= 0xbf) {
-          SDL_BlitSurface(trs_box[expanded][character - 0x80], &srcRect, screen, &dstRect);
-          goto add;
-        }
-        break;
-      }
-      /* Fall through - LNW80 II 80x24 text mode */
-    default:
-      if (trs_model == 1 && trs_clones.model == 0) {
-        /* On Model I, 0xc0-0xff is another copy of 0x80-0xbf */
-        if (character >= 0xc0)
-          character -= 0x40;
-      }
-      if (!(currentmode & INVERSE) && character >= 0x80 && character <= 0xbf) {
+  if (trs_clones.model & (CT80 | GENIE3S)) {
+    if (grafyx_overlay == 0) goto overlay;
+  } else {
+    if (trs_model == 1 && trs_clones.model == 0) {
+      /* On Model I, 0xc0-0xff is another copy of 0x80-0xbf */
+      if (character >= 0xc0)
+        character -= 0x40;
+    }
+    if (!(currentmode & INVERSE)) {
+      if (currentmode & REVERSE) expanded += 2;
+      if (character >= 0x80 && character <= 0xbf) {
         SDL_BlitSurface(trs_box[expanded][character - 0x80], &srcRect, screen, &dstRect);
         goto add;
-      } else {
-        if (trs_model > 1 && character >= 0xc0 &&
-            (currentmode & (ALTERNATE + INVERSE)) == 0) {
-          character -= 0x40;
-        }
-        if ((currentmode & INVERSE) && (character & 0x80)) {
-          expanded += 2;
-          character &= 0x7f;
-        }
       }
+    } else {
+      if (trs_model > 1 && character >= 0xc0 &&
+          (currentmode & (ALTERNATE + INVERSE)) == 0) {
+        character -= 0x40;
+      }
+      if ((currentmode & INVERSE) && (character & 0x80)) {
+        expanded += 2;
+        character &= 0x7f;
+      }
+    }
   }
   SDL_BlitSurface(trs_char[expanded][character], &srcRect, screen, &dstRect);
 
@@ -2698,13 +2693,16 @@ hrg_onoff(int enable)
 {
   if (hrg_enable == enable) return; /* State does not change. */
 
-  hrg_enable = enable;
-  grafyx_enable = enable;
-  grafyx_overlay = enable;
+  hrg_enable = grafyx_enable = grafyx_overlay = enable;
 
   if (speedup > 4) {
-    if (hrg_enable) memset(grafyx_unscaled, 0, G_YSIZE * G_XSIZE);
-    trs_screen_init(1);
+    if (hrg_enable)
+      memset(grafyx_unscaled, 0, G_YSIZE * G_XSIZE);
+
+    if (text80x24)
+      trs_screen_80x24(0);
+    else
+      trs_screen_init(1);
   } else
     trs_screen_refresh();
 }
@@ -2787,7 +2785,12 @@ void m6845_cursor(unsigned int position, int start, int end, int visible)
   if (expanded)
     srcRect.w *= 2;
 
-  expanded += (currentmode & INVERSE) && (cur_char & 0x80) ? 0 : 2;
+  if (!((currentmode & INVERSE) && (cur_char & 0x80)) &&
+      !(currentmode & REVERSE))
+    expanded += 2;
+
+  if (currentmode & INVERSE)
+    cur_char &= 0x7F;
 
   SDL_BlitSurface(trs_char[expanded][cur_char], &srcRect, screen, &dstRect);
   addToDrawList(&dstRect);
@@ -2813,10 +2816,6 @@ void m6845_screen(int chars, int lines, int raster, int factor)
     if (screen_chars < SCREEN_SIZE)
       memset(&trs_screen[screen_chars], ' ', SCREEN_SIZE - screen_chars - 1);
 
-    screen_chars = cols * rows;
-    if (screen_chars > SCREEN_SIZE)
-      screen_chars = SCREEN_SIZE;
-
     trs_screen_init(1);
   }
 }
@@ -3048,6 +3047,8 @@ void trs_sdl_init(void)
                             SDL_WINDOW_SHOWN);
   if (window == NULL)
     fatal("SDL_CreateWindow failed: %s", SDL_GetError());
+
+  SDL_StartTextInput();
 #else
   /* For NumLock on SDL >= 1.2.14 */
   putenv("SDL_DISABLE_LOCK_KEYS=1");
